import click
from godata import load_project
from godata.project import GodataProjectError

from heinlein import api


@click.command()
@click.argument("dataset_name")
@click.argument("data_type")
@click.argument("path")
@click.option("force", "-f", "--force", is_flag=True)
def add(dataset_name, data_type, path, force=False) -> bool:
    """
    Add a location on disk to a dataset
    """
    api.add(dataset_name, data_type, path, force)
    return True


@click.command()
@click.argument("dataset_name")
@click.argument("data_type")
def remove(dataset_name, data_type):
    api.remove(dataset_name, data_type)
    return True


@click.command()
@click.argument("dataset_name")
@click.argument("data_type")
def get(dataset_name, data_type) -> bool:
    """
    Get the path to a specific data type in a specific datset
    """
    try:
        project = load_project(dataset_name, ".heinlein")
    except GodataProjectError:
        print(f"Error: dataset {dataset_name} does not exist!")
        return False
    try:
        path = project.get("data/" + data_type, as_path=True)
    except GodataProjectError:
        print(f"Error: data type {data_type} does not exist in dataset {dataset_name}")
        return False
    click.echo(path)



import functools
import multiprocessing as mp
from math import ceil
from pathlib import Path

import numpy as np
import pandas as pd
import portalocker
from astropy.coordinates import SkyCoord


def warning_prompt(warning: str, options: list) -> str:
    print(warning)
    keys = [opt[0].upper() for opt in options]
    while True:
        for index, option in enumerate(options):
            print(f"{option} ({keys[index]})")
        i = input("?: ")
        if i.upper() in keys:
            return i.upper()
        else:
            print("Invalid option")


def warning_prompt_tf(warning: str) -> bool:
    options = ["Yes", "No"]
    if warning_prompt(warning, options) == "Y":
        return True
    return False


def split(path: Path, key: str, threads=1) -> None:
    files = [f for f in path.glob("*.csv") if not f.name.startswith(".")]
    print(f"Found {len(files)} files to split")
    func = functools.partial(_split, key=key)
    with mp.Pool(threads) as p:
        p.map(func, files)


def _split(file_path: Path, key: str):
    print(f"Working on file {file_path}")
    df = pd.read_csv(file_path)
    try:
        split_by_data = df[key]
    except KeyError:
        print(f"File {file_path.name} has no column {key}, skipping...")
        return
    unique_names = split_by_data.unique()
    if len(unique_names) == len(split_by_data):
        print(f"Error: key {key} seems to be unique to all items!")
        return
    sub_dfs = {name: df[df[key] == name] for name in unique_names}
    output_path = file_path.parents[0]
    print(f"Writing outputs for file {file_path}")
    write_split_output_file(sub_dfs, output_path)


def write_split_output_file(dfs, output):
    for key, df in dfs.items():
        fpath = output / ".".join([str(key), "csv"])
        try:
            first = pd.read_csv(fpath)
            output_df = pd.concat([first, df])
        except FileNotFoundError:
            output_df = df

        with portalocker.Lock(fpath, "w") as of:
            output_df.to_csv(of, index=False)


def initialize_grid(bounds, area, density):
    ra1, ra2 = bounds[0], bounds[2]
    dec1, dec2 = bounds[1], bounds[3]
    dra = ra2 - ra1
    ddec = dec2 - dec1
    npoints = round(density * area)
    ratio = ddec / dra
    nra = np.sqrt(npoints / ratio)
    ndec = nra * ratio
    nra = ceil(nra) + 1
    ndec = ceil(ndec) + 1
    dra_ = dra / (nra + 1)
    ddec_ = ddec / (ndec + 1)
    gra = [ra1 + i * dra_ for i in range(1, nra + 1)]
    gdec = [dec1 + i * ddec_ for i in range(1, ndec + 1)]
    x_, y_ = np.meshgrid(gra, gdec)
    cpoints = np.vstack([x_.ravel(), y_.ravel()])
    coords = SkyCoord(cpoints[0], cpoints[1], unit="deg")
    return coords



import numpy as np
from astropy.io import ascii
from astropy.table import Table
from godata.project import GodataProject, GodataProjectError
from sqlalchemy import create_engine, text

from heinlein.dtypes.catalog import Catalog
from heinlein.dtypes.handlers import handler


class heinleinIoException(Exception):
    pass


def get_catalog_handler(project: GodataProject, dconfig: dict):
    try:
        path = project.get("data/catalog")
    except GodataProjectError:
        _ = project.list("data/catalog").get
        return CsvCatalogHandler(project, dconfig)
    if path.suffix == ".db":
        return SQLiteCatalogHandler(project, dconfig)


class CsvCatalogHandler(handler.Handler):
    def __init__(self, project: GodataProject, config: dict, *args, **kwargs):
        super().__init__(project, config, "catalog")

    def get_data(self, region_names: list, *args, **kwargs):
        """
        Default handler for a catalog.
        Loads a single catalog, assuming the region name can be found in the file name.
        """
        storage = {}
        known_files = self._project.list("data/catalog")
        files = {}
        for name in region_names:
            region_file = list(filter(lambda x: name in x, known_files["files"]))
            if len(region_file) != 1:
                raise heinleinIoException(f"Multiple files found for region {name}! ")
            path = self._project.get("data/catalog/" + region_file[0], as_path=True)
            files.update({name: path})

        for name, path in files.items():
            if not path.exists():
                print(
                    f"Path {self._path} does not exist! Perhaps it is in an external"
                    " storage device that isn't attached."
                )
                return None
            data = ascii.read(path)
            storage.update({name: Catalog(data, self._config)})
        return storage


class SQLiteCatalogHandler(handler.Handler):
    def __init__(self, project, config: dict, *args, **kwargs):
        super().__init__(project, config, "catalog")
        try:
            self.db_path = self._project.get("data/catalog")
        except GodataProjectError:
            raise heinleinIoException(
                "SQLite catalog handler requires SQLite "
                "database, but found a folder!"
            )

        self._create_engine()

    def _create_engine(self, *args, **kwargs):
        self._engine = create_engine(f"sqlite:///{self.db_path}")
        self._con = self._engine.connect()
        sql = "SELECT * FROM sqlite_master where type='table'"
        cur = self.execute_query(sql)
        self._tnames = [t[1] for t in cur.fetchall()]

    def get_data(self, region_names: list, *args, **kwargs):
        subregion_key = self._config.get("subregion", None)
        if subregion_key is not None:
            splits = [rname.split(".") for rname in region_names]
            regions_to_get = {}
            for split in splits:
                if len(split) == 2:
                    if split[0] in regions_to_get.keys():
                        regions_to_get[split[0]].append(split[1])
                    else:
                        regions_to_get.update({split[0]: [split[1]]})
                elif len(split) == 2 and split[0] not in regions_to_get.keys():
                    regions_to_get.update({split[0]: []})

            storage = self.get_with_subregions(regions_to_get)
        else:
            storage = self._get(region_names)

        return {k: Catalog(table) for k, table in storage.items()}

    def get_with_subregions(self, regions: dict):
        subregion_key = self._config.get("subregion", None)
        region_key = self._config.get("region", None)
        storage = {}
        for region, subregions in regions.items():
            region_names = [".".join([region, sr]) for sr in subregions]
            if str(region) in self._tnames:
                table = self.get_where(region, {subregion_key: subregions})
                if len(table) != 0:
                    for index, sr in enumerate(subregions):
                        mask = (table[region_key].astype(str) == region) & (
                            table[subregion_key].astype(str) == sr
                        )
                        storage.update({region_names[index]: table[mask]})
                else:
                    storage.update({sr: Catalog() for sr in region_names})
                    # This needs to be fixed
            elif len(self._tnames) == 1:
                region_names = [".".join([region, sr]) for sr in subregions]

                table = self.get_where(
                    self._tnames[0],
                    {region_key: region.name, subregion_key: subregions},
                )
                for index, sr in enumerate(subregions):
                    mask = table[subregion_key] == sr
                    storage.update({region_names[index]: table[mask]})
            else:
                storage.update({sr: Table() for sr in region_names})
        return storage

    def _get(self, region_names: list):
        storage = {}
        for region in region_names:
            if region in self._tnames:
                table = self.get_all(region)

                storage.update({region: table})

            elif len(self._tnames) == 1:
                region_key = self._config.get("region", None)
                table = self.get_where(self._tnames[0], {region_key: region.name})
                storage.update({region: table})
        return storage

    def get_where(self, tname: str, conditions: dict):
        base_query = f'SELECT * FROM "{tname}" WHERE '
        base_condition = '{} = "{}"'
        multiple_base_conditions = "{} IN ({})"
        output_conditions = []
        for k, vs in conditions.items():
            if isinstance(vs, list):
                c = '","'.join([str(v) for v in vs])
                c = '"' + c + '"'
                output_conditions.append(multiple_base_conditions.format(k, c))
            else:
                output_conditions.append(base_condition.format(k, vs))
        query = base_query + " AND ".join(output_conditions)
        table = self.execute_query(query)
        return self._parse_return(table)

    def get_all(self, tname):
        query = f'SELECT * FROM "{tname}"'
        table = self.execute_query(query)
        return self._parse_return(table)

    def execute_query(self, query):
        q = text(query)
        cur = self._con.execute(q)
        return cur

    def _parse_return(self, cursor, *args, **kwargs):
        rows = cursor.fetchall()
        if len(rows) == 0:
            return Catalog()
        rows = np.array(rows, dtype=object)
        # replace all None with np.nan
        rows[rows is None] = np.nan
        columns = cursor.keys()
        data = Table(rows=rows, names=columns)
        return data



from __future__ import annotations

import json
import logging
from abc import ABC, abstractmethod
from functools import singledispatchmethod
from typing import Any

from astropy.coordinates import SkyCoord
from shapely.geometry import Polygon
from spherical_geometry.vector import lonlat_to_vector, vector_to_lonlat

from heinlein.locations import MAIN_CONFIG_DIR

logger = logging.getLogger("region")


def load_config(*args, **kwargs):
    """
    Loads the region config.
    """
    config_location = MAIN_CONFIG_DIR / "region" / "region.json"
    with open(config_location, "r") as f:
        config = json.load(f)
        return config


current_config = load_config()


class BaseRegion(ABC):
    _config = current_config

    def __init__(self, geometry, type, name=None, *args, **kwargs):
        """
        Base region object.
        Regions should always be initialized with heinlein.Region

        parameters:

        geometry: <spherical_geometry.SingleSphericalPolygon> The sky region
        type: <str> The type of the region
        name: <str> The name of the region (optional)
        """
        self._spherical_geometry = geometry
        self._type = type
        self.name = name
        self.setup()

    def setup(self, *args, **kwargs):
        """
        Perform setup for the region
        """

        if not hasattr(self, "_flat_geometry"):
            points = self._spherical_geometry.points
            self._flat_geometry = Polygon(points)

        if not hasattr(self, "_flat_sky_geometry"):
            points = self._spherical_geometry.points
            v = vector_to_lonlat(points[:, 0], points[:, 1], points[:, 2])
            ra = [r.round(2) for r in v[0]]
            dec = [d.round(2) for d in v[1]]
            self._flat_sky_geometry = Polygon(list(zip(ra, dec)))

    def __setstate__(self, state):
        """
        Fixes reading pickled regions
        """
        self.__dict__.update(state)
        self._config = current_config
        self.setup()

    def __getattr__(self, __name: str) -> Any:
        """
        Implements geometry relations for regions. Delegates unknown methods to the
        underlying spherical geometry object, IF that method is explicitly permitted
        in heinlein/config/region/region.json
        """
        try:
            cmd_name = self._config["allowed_predicates"][__name]
            attr = getattr(self.spherical_geometry, cmd_name)

            def func(x):
                return attr(x.spherical_geometry)

            return func
        except KeyError:
            raise AttributeError(f"{self._type} has no attribute '{__name}'")

    def _delegate_relationship(
        self, other: BaseRegion, method_name: str, *args, **kwargs
    ) -> Any:
        attr = getattr(self._geometry, method_name)
        return attr(other)

    @property
    def geometry(self, *args, **kwargs):
        """
        Returns the flattened 3D geometry for the object
        """
        return self._flat_geometry

    @property
    def sky_geometry(self, *args, **kwargs):
        """
        Returns the flat geometry, using sky coordinates
        """
        return self._flat_sky_geometry

    @property
    def spherical_geometry(self):
        """
        Returns the correct spherical geometry for the object
        """
        return self._spherical_geometry

    @property
    def type(self) -> str:
        """
        Returns the type of the region
        """
        return self._type

    def center(self, *args, **kwargs):
        pass

    def get_grid(self, density, *args, **kwargs):
        try:
            return self._grids[density]

        except KeyError:
            g = self.initialize_grid(density)
            self._grids[density] = g
            return g

        except AttributeError:
            self._grids = {}
            g = self.initialize_grid(density)
            self._grids[density] = g
            return g

    @singledispatchmethod
    def contains(self, point, *args, **kwargs):
        return self._spherical_geometry.contains_point(point)

    @contains.register
    def _(self, point: SkyCoord):
        vec = lonlat_to_vector(point.ra, point.dec)
        return self.contains(vec)

    @abstractmethod
    def initialize_grid(self, density, *args, **kwargs):
        pass



from pathlib import Path

from godata import load_project
from godata.project import GodataProjectError

from heinlein.manager import initialize_dataset
from heinlein.utilities import warning_prompt_tf

"""
Backend API functions
"""


def add(name, dtype, path, force=False, *args, **kwargs) -> bool:
    """
    Add a location on disk to a dataset
    """
    cwd = Path.cwd()

    if path != "cwd":
        path = cwd / path
    else:
        path = cwd
    if not path.exists():
        print(f"Error: {path} not found!")
        return

    try:
        project = load_project(name, ".heinlein")
    except GodataProjectError:
        if not force:
            write_new = warning_prompt_tf(
                f"Survey {name} not found, would you like to initialize it? "
            )
            if not write_new:
                print("Aborting...")
                return
        initialize_dataset(name, path)
        project = load_project(name, ".heinlein")

    project.link(path, "data/" + dtype, overwrite=force, recursive=True)
    return True


def remove(name: str, dtype: str):
    """
    Remove a datatype from a dataset
    """
    try:
        project = load_project(name, ".heinlein")
    except GodataProjectError:
        print(f"Error: dataset {name} does not exist!")
        return True

    project.remove("data/" + dtype)
    return True



import json
import logging
import multiprocessing as mp
from importlib import import_module
from inspect import getmembers, isclass, isfunction
from itertools import chain
from pathlib import Path
from types import ModuleType

from cacheout import LRUCache
from godata import create_project, has_collection, has_project, load_project
from godata.project import GodataProjectError

from heinlein.config.config import globalConfig
from heinlein.locations import BASE_DATASET_CONFIG_DIR
from heinlein.region.base import BaseRegion
from heinlein.utilities import warning_prompt, warning_prompt_tf

logger = logging.getLogger("manager")


known_datasets = ["des", "cfht", "hsc", "ms"]

active_managers = {}


def get_manager(name):
    try:
        am = active_managers[name]
        return am
    except KeyError:
        mgr = DataManager(name)
        active_managers.update({name: mgr})
        return mgr


def initialize_dataset(name: str, *args, **kwargs):
    """
    Initializes a new dataset. Checks to see if a custom implementation exists for a
    given dataset. These have to be installed separately from the main package.
    """
    ext = get_external_implementation(name)
    config_data = ext.load_config()

    project = create_project(name, ".heinlein")
    print(f"Initializing dataset {name} with config {config_data}")
    project.store(config_data, "meta/config")


def get_external_implementation(name: str) -> ModuleType:
    """
    Checks to see if a custom implementation exists for a given dataset. These have
    to be installed separately from the main package. Prompts the user to install
    the package if it is not found, but is known.
    """
    if name not in known_datasets:
        raise ValueError(f"Dataset {name} is not supported!")
    else:
        # try to import it
        module_name = "heinlein_" + name
        try:
            module = import_module(module_name)
        except ImportError:
            print(
                f"Dataset `{name}` is a known dataset, but needs to be installed"
                f" separately. You can install it with `pip install heinlein[{name}]`."
            )
            exit()
        return getattr(module, name)


def check_overload(f):
    def wrapper(self, *args, **kwargs):
        bypass = kwargs.get("bypass", False)
        if self.external is None or bypass:
            return f(self, *args, **kwargs)
        else:
            try:
                ext_fn = self._external_definitions[f.__name__]
                return ext_fn(self, *args, **kwargs)
            except KeyError:
                return f(self, *args, **kwargs)

    return wrapper


def get_default_config() -> Path:
    """
    Returns the default config for a dataset
    """
    path = BASE_DATASET_CONFIG_DIR / "default.json"
    with open(path, "rb") as f:
        data = json.load(f)

    return data


logger = logging.getLogger("manager")


class DataManager:
    def __init__(self, name, *args, **kwargs):
        """
        The datamanger is a fairly thin wrapper around godata to track where data is
        and store configuration.

        parameters:

        name: <str> The name of the dataset
        """
        self.name = name
        self.globalConfig = globalConfig
        self._setup()
        self._cache = {}
        self._cache_lock = mp.Lock()

    def get_config(self):
        return self.config.get("meta/config")

    def _setup(self, *args, **kwargs) -> None:
        """
        Performs basic setup of the manager
        Loads datset config if it exists, or prompts
        user if dataset does not exist.
        """
        try:
            # Find the external implementation for this dataset, if it exists.
            self.external = get_external_implementation(self.name)
        except KeyError:
            self.external = None
        self._initialize_external_implementation()

        if not has_collection(".heinlein") or not has_project(
            self.name, ".heinlein"
        ):  # If this dataset does not exist
            if self.globalConfig.interactive:
                write_new = warning_prompt_tf(
                    f"Survey {self.name} not found, would you like to initialize it? "
                )
                if write_new:
                    if self.external is not None:
                        config_data = self.external.get_config()
                    else:
                        config_data = get_default_config()
                    self.config = create_project(self.name, ".heinlein")
                    self.config.store(config_data, "config")

                else:
                    self.ready = False
            else:
                raise OSError(f"Dataset {self.name} does not exist!")

        else:  # If it DOES exist, get the config data
            self.config = load_project(self.name, ".heinlein")

    def _initialize_external_implementation(self):
        if self.external is None:
            self._external_definitions = {}
            return
        fns = [f for f in getmembers(self.external, isfunction)]
        fns = list(
            filter(lambda f, m=self.external: f[1].__module__ == m.__name__, fns)
        )
        external_functions = {fn[0]: fn[1] for fn in fns}

        classes = [f for f in getmembers(self.external, isclass)]
        classes = list(
            filter(lambda f, m=self.external: f[1].__module__ == m.__name__, classes)
        )
        external_classes = {cl[0]: cl[1] for cl in classes}

        fn_keys = set(external_functions.keys())
        cls_keys = set(external_classes.keys())
        if len(fn_keys.intersection(cls_keys)) != 0:
            print(
                "Error: Overloaded functions and classes in dataset implementations"
                "Should all have unique names, but found duplicates for {keys_}"
            )
            exit()
        self._external_definitions = {**external_classes, **external_functions}

    def get_external(self, key, *args, **kwargs):
        return self._external_definitions.get(key, None)

    def get_path(self, dtype: str, *args, **kwargs):
        return self.config.get(f"data/{dtype}", as_path=True)

    def load_handlers(self, *args, **kwargs):
        from heinlein.dtypes import handlers

        if not hasattr(self, "_handlers"):
            known_dtypes_ = self.config.list("data")
            known_dtypes = []
            for ff in known_dtypes_.values():
                known_dtypes.extend(ff)
            self._handlers = handlers.get_file_handlers(
                known_dtypes, self.config, self._external_definitions
            )

    @staticmethod
    def exists(name: str) -> bool:
        """
        Checks if a datset exists
        """
        try:
            return has_project(name, ".heinlein")
        except GodataProjectError:  # Need to eport the GodataProjectError here
            return False

    def add_data(self, dtype: str, path: Path, overwrite=False) -> bool:
        """
        Add data to a datset. Note that this only gives the manager
        a path to the data. The manager itself does not know what kind
        of data it is or how to use it. Usually this will be invoked
        by a command line script.

        Params:

        dtype <str>: Type of data being added (i.e. "catalog")
        path <pathlib.Path>: Path to the data

        Returns:

        bool: Whether or not the file was sucessfully added
        """
        if not self.ready:
            return False

        has_dtype_already = False
        has_any_data = "data" in self.config.list()["folders"]
        if has_any_data:
            known_dtypes = chain.from_iterable(self.config.list("data").values())
            has_dtype_already = dtype in known_dtypes

        if has_dtype_already and not overwrite:
            msg = f"Datatype {dtype} already found for survey {self.name}."
            options = ["Overwrite", "Merge", "Abort"]
            choice = warning_prompt(msg, options)
            if choice == "A":
                return False
            elif choice == "M":
                raise NotImplementedError

        if path.is_dir():
            print(
                "Adding folder to the dataset recursively. This may take some"
                " time if there are a lot of folders or sub-folders."
            )
        self.config.link(path, f"data/{dtype}", recursive=True)
        return True

    def remove_data(self, dtype: str) -> bool:
        """
        Remove data from a datset. Usually this will be invoked
        by a command line script.

        Params:

        dtype <str>: Type of data being added (i.e. "catalog")

        Returns:

        bool: Whether or not the file was sucessfully removed
        """
        if dtype not in self.config.list("data")["files"]:
            print(f"Error: dataset {self.name} has no data of type {dtype}")
        self.config.remove(f"data/{dtype}", recursive=True)

    @check_overload
    def get_from(self, dtypes: list, region_overlaps: list, *args, **kwargs):
        return_types = []
        new_data = {}
        if not isinstance(region_overlaps[0], str):
            regnames = [r.name for r in region_overlaps]
        else:
            regnames = region_overlaps

        self.load_handlers()

        for dtype in dtypes:
            try:
                path = self.config.has_path(f"data/{dtype}")
                return_types.append(dtype)
            except KeyError:
                print(f"No data of type {dtype} found for dataset {self.name}!")
        cached = self.get_cached_values(dtypes, regnames)
        for dtype in return_types:
            if dtype in cached.keys():
                dtype_cache = cached[dtype]
                regions_to_get = [r for r in regnames if r not in dtype_cache.keys()]
            else:
                regions_to_get = regnames
            if len(regions_to_get) != 0:
                data_ = self._handlers[dtype].get_data(regions_to_get, *args, **kwargs)
                new_data.update({dtype: data_})

        if len(new_data) != 0:
            self.cache(new_data)

        storage = {}
        keys = set(new_data.keys()).union(set(cached.keys()))

        for k in keys:
            data = cached.get(k, {})
            new_d = new_data.get(k, {})
            if new_d is None and data is None:
                path = self.get_path(k)
                storage.update({k: path})
            elif not isinstance(new_d, dict):
                storage.update({k: new_d})
            else:
                data.update(new_d)
                storage.update({k: data})
        return storage

    @check_overload
    def get_data(
        self,
        dtypes: list,
        query_region: BaseRegion,
        region_overlaps: list,
        *args,
        **kwargs,
    ) -> dict:
        """
        Get data of a specificed type
        The manager is responsible for finding the path, and the giving
        it to the handlers
        """

        storage = self.get_from(dtypes, region_overlaps, *args, **kwargs)
        storage = self.parse_data(storage, *args, **kwargs)
        return storage

    def load(self, regions: list, dtypes: list, *args, **kwargs):
        """
        Loads data for particular named regions into the cache
        """
        self.get_from(dtypes, regions)

    def dump(self, regions: list, *args, **kwargs):
        """
        Dumps data for some particular named regions from the cache
        """
        for dtype, data in self._cache.items():
            nd = data.delete_many(regions)
            logging.info(f"Delete {nd} items from the {dtype} cache")

    def dump_all(self):
        for dtype, data in self._cache.items():
            nd = data.clear()
            logging.info(f"Delete {nd} items from the {dtype} cache")

    def parse_data(self, data, *args, **kwargs):
        return_data = {}
        for dtype, values in data.items():
            # Now, we process into useful objects and filter further
            if data is None:
                logger.error(f"Unable to find data of type {dtype}")
                continue
            try:
                objs = list(values.values())
                data_obj = objs[0].combine(objs)
                return_data.update({dtype: data_obj})
            except IndexError:
                return_data.update({dtype: None})
        return return_data

    def get_cached_values(self, dtypes: list, region_overlaps: list):
        cached_values = {}
        with self._cache_lock:
            for dtype in dtypes:
                try:
                    dtype_cache = self._cache[dtype]
                except KeyError:
                    # No cache for this datatype, continue
                    continue
                cached = dtype_cache.get_many(region_overlaps)
                cached_values.update({dtype: cached})
        return cached_values

    def cache(self, data_storage: dict):
        """
        Top-level modules think in terms of datatypes, but the cache thinks in terms
        of regions so we have to do a translation

        """
        with self._cache_lock:
            for dtype, data in data_storage.items():
                if (data is None) or not isinstance(data, dict):
                    continue
                try:
                    dtype_cache = self._cache[dtype]
                except KeyError:
                    dtype_cache = LRUCache(maxsize=0)
                    self._cache.update({dtype: dtype_cache})
                dtype_cache.add_many(
                    {reg_name: d_obj for reg_name, d_obj in data.items()}
                )
                self._cache.update({dtype: dtype_cache})

    def get_handler(self, dtype: str, *args, **kwargs):
        pass

    def clear_all_data(self, *args, **kwargs) -> None:
        self.config.remove("data", recursive=True)



import astropy.units as u
from shapely import GeometryCollection, STRtree, union_all

from .region import BaseRegion
from .sampling import Sampler


def build_tree(regions: list[BaseRegion]):
    """
    Builds an STRtree from a list of regions
    """
    geos = [r.geometry for r in regions]
    tree = STRtree(geos)
    return tree


def build_footprint(regions: list[BaseRegion]) -> GeometryCollection:
    """
    Stitch together a list of regions into a single footprint
    """
    shapely_geos = [r._flat_sky_geometry for r in regions]
    collection = union_all(shapely_geos)
    return collection


class Footprint:
    """
    The footprint class is a container for a set of regions. It is used to represent
    the geometry of the survey, which is used to accelerate queries.
    """

    def __init__(self, regions: dict[str, BaseRegion], *args, **kwargs):
        self._regnames = list(regions.keys())
        self._regions = list(regions.values())
        self._tree = build_tree(self._regions)
        self._footprint = build_footprint(self._regions)
        self._sampler = Sampler(self._footprint)

    def get_overlapping_regions(self, region: BaseRegion):
        """
        Returns a list of regions that overlap with the given region
        """
        overlap_idx = self._tree.query(region.geometry)
        overlaps = [self._regions[i] for i in overlap_idx]
        overlaps = filter(lambda x: x.intersects(region), overlaps)
        return list(overlaps)

    def get_overlapping_region_names(self, region: BaseRegion):
        """
        Returns a list of region names that overlap with the given region
        """
        return [r.name for r in self.get_overlapping_regions(region)]

    def sample(self, n: int = 1, tolerance: u.Quantity = None, *args, **kwargs):
        """
        Return n random points from the footprint
        """
        if n == 1:
            return self._sampler.sample(tolerance=tolerance)
        else:
            return [self._sampler.sample(tolerance=tolerance) for _ in range(n)]



import logging
from functools import partial
from inspect import getmembers
from typing import Union

import astropy.units as u

from heinlein.manager import get_manager
from heinlein.manager.manager import DataManager
from heinlein.region import BaseRegion, Region
from heinlein.region.footprint import Footprint

logger = logging.getLogger("Dataset")


def check_overload(f):
    def wrapper(self, *args, **kwargs):
        overload = self.manager.get_external(f.__name__)
        bypass = kwargs.get("bypass", False)
        if bypass or overload is None:
            return f(self, *args, **kwargs)
        else:
            return overload(self, *args, **kwargs)

    return wrapper


class dataset_extension:
    def __init__(self, function):
        self.function = function
        self.extension = True
        self.name = function.__name__

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)


class Dataset:
    """
    The Dataset class is the core of heinlein's user interface. It contains routines
    for querying data from the underlying dataset.

    The dataset class should not be created directly. Instead use "load_dataset."

    """

    def __init__(self, manager: DataManager, *args, **kwargs):
        self.manager = manager
        self.config = manager.get_config()
        self._extensions = {}
        self._parameters = {}
        self._setup()

    def __getattr__(self, key__):
        """
        I went through a phase where I thought defining magic methods all over the place
        was really cool. I've since learned that it sometimes creates more complexity
        than it is worth, but this is one case where it makes sense.

        There are certain methods that need to be defined or overloaded for specific
        datasets. This method allows that to be largely transparent to the end user,
        without cluttering the main class with a bunch of extra methods.
        """

        try:
            f = self._extensions[key__]
            return partial(f, self)
        except KeyError:
            raise AttributeError(
                f"{type(self).__name__} object has no attribute '{key__}'"
            )

    @property
    def name(self):
        return self.manager.name

    def _setup(self, *args, **kwargs) -> None:
        """
        Searches for an external implementation of the datset
        This is used for datasets with specific needs (i.e. specific surveys)
        """
        external = self.config["implementation"]

        if external and self.manager.external is None:
            raise NotImplementedError(
                f"No implementation code found for datset {self.name}"
            )

        try:
            get_regions = self.manager.get_external("load_regions")
        except KeyError:
            raise NotImplementedError(
                f"Dataset {self.name} does not have a setup method!"
            )

        regions = get_regions()
        self.footprint = Footprint(regions)
        self._load_extensions()

    def set_parameter(self, name, value):
        """
        Parameters are used to store metadata that may be useful to plugins etc.
        """
        self._parameters.update({name: value})

    def get_parameter(self, name):
        return self._parameters.get(name, None)

    def _load_extensions(self, *args, **kwargs):
        """
        Loads extensions for the particular dataset. These are defined externally
        """
        ext_objs = list(
            filter(
                lambda f: type(f[1]) == dataset_extension,
                getmembers(self.manager.external),
            )
        )
        self._extensions.update({f[0]: f[1] for f in ext_objs})

    def get_path(self, dtype: str, *args, **kwargs):
        """
        Gets the path to where a particular item in a dataset is stored on disk
        """
        return self.manager.get_path(dtype)

    def sample_generator(
        self,
        samples,
        sample_type="cone",
        sample_dimensions=45 * u.arcsec,
        dtypes=["catalog"],
        *args,
        **kwargs,
    ):
        """
        Often, we want to get many many samples in a row. The problem here
        is that the cache will very quickly blow up if the samples cover may
        of the survey's subregions. This methord returns a generator that
        yields samples from the survey. It orders them such that it can
        only load a few subregions at a time, then dumps them from the cache when
        they are done.
        """
        if sample_type != "cone":
            raise NotImplementedError("Only cone sampling is currently supported")

        samples = [Region.circle(center=s, radius=sample_dimensions) for s in samples]

        overlaps = self.footprint.get_overlapping_region_names(samples)
        partitions = {}

        for i, sample in enumerate(samples):
            overlap = overlaps[i]
            if len(overlap) == 1:
                okey = overlap[0]
            else:
                overlap.sort()
                okey = "/".join(overlap)

            if okey in partitions.keys():
                partitions[okey].append(sample)
            else:
                partitions[okey] = [sample]

        samples = partitions
        counts = [p.count("/") + 1 for p in samples.keys()]
        singles = []

        for i, (regs, s) in enumerate(samples.items()):
            if counts[i] == 1:
                continue
            regs_to_get = regs.split("/")
            for s_ in s:
                yield (s_, self.get_data_from_region(s_, dtypes))
            for reg_ in regs_to_get:
                # Now, get the samples that fall into a single one of the regions
                if reg_ in singles:
                    continue
                try:
                    samples_in_reg = samples[reg_]
                except KeyError:
                    continue
                singles.append(reg_)
                for s_ in samples_in_reg:
                    yield (s_, self.get_data_from_region(s_, dtypes))
            # Now, we dump the data from those regions
            self.dump_all()
        # Now we go through all the samples that fall in a single survey region
        # that were NOT covered before
        for i, (reg, s) in enumerate(samples.items()):
            if counts[i] != 1 or reg in singles:
                continue
            else:
                for s_ in s:
                    yield (s_, self.get_data_from_region(s_, dtypes))
                self.dump_all()

    @check_overload
    def get_data_from_named_region(
        self, name: str, dtypes: Union[str, list] = "catalog"
    ):
        """
        Given a region name, returns the data of type dtypes in that region.
        """
        if name not in self._region_names:
            print(f"Unable to find region named {name} in dataset {self.name}")
            return

        regs_ = self._regions[self._region_names == name]
        return self.manager.get_from(dtypes, regs_)

    def load(self, regions, dtypes, *args, **kwargs):
        """
        Pre-loads some regions into the cache.
        """
        if isinstance(regions, str):
            regions = [regions]
        if not all([r in self._region_names for r in regions]):
            logging.error("Regions not found!")
            return
        self.manager.load(regions, dtypes)

    def dump(self, regions, *args, **kwargs):
        """
        Dumps some regions from the cache.
        """
        if isinstance(regions, str):
            regions = [regions]
        self.manager.dump(regions)

    def dump_all(self):
        self.manager.dump_all()

    def get_data_from_region(
        self,
        query_region: BaseRegion,
        dtypes: Union[str, list] = "catalog",
        *args,
        **kwargs,
    ) -> dict:
        """
        Get data of type dtypes from a particular region

        Paramaters:

        region <BaseRegion> heinlein Region object
        dtypes <str> or <list>: list of data types to return

        """
        method = self.manager.get_external("get_overlapping_regions")
        if method is not None:
            overlaps = method(self, query_region)
        else:
            overlaps = self.footprint.get_overlapping_regions(query_region)
        if len(overlaps) == 0:
            print("Error: No objects found in this region!")
            return
        data = {}
        if isinstance(dtypes, str):
            dtypes = [dtypes]

        data = self.manager.get_data(dtypes, query_region, overlaps, *args, **kwargs)
        return_data = {}
        for dtype, obj_ in data.items():
            try:
                return_data.update({dtype: obj_.get_data_from_region(query_region)})
            except AttributeError:
                return_data.update({dtype: obj_})

        for dtype, d_ in return_data.items():
            if len(d_) == 0:
                continue
            try:
                aliases = self._aliases
            except AttributeError:
                aliases = {}
                self._aliases = aliases
            try:
                alias = aliases[dtype]
                d_.add_aliases(alias)
            except KeyError:
                continue

        return return_data

    def cone_search(self, center, radius, *args, **kwargs):
        """
        A convinience method for doing a cone search. Basically just
        constructs a circular region and calls get_data_from_region.
        """
        reg = Region.circle(center=center, radius=radius)
        return self.get_data_from_region(reg, *args, **kwargs)


def load_dataset(name: str) -> Dataset:
    manager = get_manager(name)
    ds = Dataset(manager)
    return ds


def load_current_config(name: str):
    manager = get_manager(name)
    return manager.config



from dynaconf import Dynaconf

from heinlein.locations import MAIN_CONFIG_DIR

globalConfig = Dynaconf(
    envvar_prefix="DYNACONF",
    settings_files=[MAIN_CONFIG_DIR / "config.json", ".secrets.json"],
)

# `envvar_prefix` = export envvars with `export DYNACONF_FOO=bar`.
# `settings_files` = Load these files in the order.



from functools import singledispatchmethod
from typing import Union

import astropy.units as u
from astropy.coordinates import SkyCoord
from shapely import geometry
from shapely.geometry import Point
from shapely.geometry.base import BaseGeometry
from spherical_geometry.polygon import SingleSphericalPolygon
from spherical_geometry.vector import vector_to_lonlat

from heinlein.region import sampling
from heinlein.region.base import BaseRegion
from heinlein.utilities.utilities import initialize_grid


class Region:
    @staticmethod
    def circle(
        center: Union[SkyCoord, tuple],
        radius: Union[u.Quantity, float],
        *args,
        **kwargs
    ):
        """
        Return a circular region. Centered on center, radius of `radius`
        The "center" is anything that can be parsed to a SkyCoord
        If no units provided, will default to degrees
        """
        sky_center = center
        sky_radius = radius
        if type(center) == SkyCoord and type(radius) == u.Quantity:
            return CircularRegion(center, radius)
        elif type(center) == tuple:
            try:
                center_coord = SkyCoord(*sky_center)
            except u.UnitTypeError:
                center_coord = SkyCoord(*sky_center, unit="deg")
        elif type(center) == SkyCoord:
            center_coord = center

        if type(sky_radius) != u.Quantity:
            sky_radius = sky_radius * u.deg

        return CircularRegion(center_coord, sky_radius, *args, **kwargs)

    @staticmethod
    def polygon(coords, *args, **kwargs):
        """
        Return a generically-shaped region:
        """
        if type(coords) == SingleSphericalPolygon:
            return PolygonRegion(coords, *args, **kwargs)
        elif issubclass(type(coords), BaseGeometry):
            points = coords.exterior.xy
            poly = SingleSphericalPolygon.from_radec(points[0], points[1])
            return PolygonRegion(poly, *args, **kwargs)

    @staticmethod
    def box(bounds, *args, **kwargs):
        if not (isinstance(bounds, list) or len(args) == 3):
            print("Error box region expects 4 inputs")
            return
        if len(args) == 3:
            bounds_ = [bounds] + list(args)
        else:
            bounds_ = bounds
        box_ = geometry.box(*bounds_)
        b_ = Region.polygon(box_)
        b_.box_ = box_
        return b_


class PolygonRegion(BaseRegion):
    def __init__(self, polygon, name: str = None, *args, **kwargs):
        """
        Basic general-shape region object.

        Parameters:

        polygon <spherical_geometry.SingleSphericalPolygon>: The spherical polygon
        representing the region

        name <str>: a name for the region (optional)
        """
        super().__init__(polygon, "PolygonRegion", name)
        self._sampler = None

    @property
    def center(self) -> Point:
        """
        Return the center of the region
        """
        return self._flat_geometry.centroid

    def generate_circular_tile(self, radius, *args, **kwargs):
        """
        Return a circular tile, drawn randomly from the region.
        """
        if getattr(self, "_sampler", None) is None:
            self._get_sampler()
        return self._sampler.get_circular_sample(radius)

    def generate_circular_tiles(self, radius, n, *args, **kwargs):
        if self._sampler is None:
            self._get_sampler()
        return self._sampler.get_circular_samples(radius, n)

    def _get_sampler(self, *args, **kwargs):
        self._sampler = sampling.Sampler(self)

    def contains(self, reg: BaseRegion):
        return self.sky_geometry.contains(reg.sky_geometry)

    def initialize_grid(self, density=1000, *args, **kwargs):
        bounds = self.sky_geometry.bounds
        area = self.sky_geometry.area
        coords = initialize_grid(bounds, area, density)
        return coords

    generate_grid = initialize_grid


class CircularRegion(BaseRegion):
    def __init__(
        self, center: SkyCoord, radius: u.Quantity, name=None, *args, **kwargs
    ) -> None:
        """
        Circular region. Accepts point-radius for initialization.

        parameters:

        center <SkyCoord>: The center of the region
        radius <astropy.units.quantity>: The radius of the region
        name <str>: a name for the region (optional)
        """

        self._skypoint = center
        self._radius = radius
        self._center = Point(center.ra.to_value("deg"), center.dec.to_value("deg"))

        geometry = SingleSphericalPolygon.from_cone(
            center.ra.to(u.deg).value,
            center.dec.to(u.deg).value,
            self._radius.to(u.deg).value,
            *args,
            **kwargs
        )
        super().__init__(geometry, "CircularRegion", name, *args, **kwargs)

    def translate(self, x: u.Quantity, y: u.Quantity, *args, **kwargs):
        """
        Translate the region by x and y
        """
        new_center = (self._skypoint.ra + x, self._skypoint.dec + y)
        return Region.circle(new_center, self.radius)

    @property
    def center(self) -> Point:
        return self._center

    @property
    def coordinate(self) -> SkyCoord:
        return self._skypoint

    @property
    def radius(self) -> u.quantity:
        return self._radius

    @singledispatchmethod
    def contains(self, point: SkyCoord):
        separation = point.separation(self.coordinate)
        return separation <= self.radius

    @contains.register
    def _(self, point: geometry.Point, *args, **kwargs):
        lonlat = vector_to_lonlat(point.x, point.y, point.z)
        return self.contains(SkyCoord(*lonlat, unit="deg"))

    def initialize_grid(self, density, *args, **kwargs):
        bounds = self.sky_geometry.bounds
        area = geometry.box(*bounds).area
        grid = initialize_grid(bounds, area, density)
        center = self.coordinate
        return grid[center.separation(grid) < self.radius]



import functools
import multiprocessing as mp
import sqlite3
from pathlib import Path

import pandas as pd

import heinlein
from heinlein import api


def split_catalog(args) -> None:
    files = [file for file in args.input_path.glob(f"*{args.input_format}")]
    files = [f for f in files if not f.name.startswith(".")]
    print(f"Found {len(files)} files to read from")
    ds = heinlein.dataset.load_current_config(args.survey_name)
    try:
        catalog_config = ds["data"]["catalog"]
    except KeyError:
        catalog_config = ds["dconfig"]["catalog"]

    region_key = catalog_config["region"]
    subregion_key = catalog_config.get("subregion", False)
    if args.threads > 1:
        mgr = mp.Manager()
        accessed = mgr.dict()
    else:
        accessed = {}

    if args.output_format == "sqlite":
        output_path = args.output / ".".join([args.survey_name, "db"])
        output_f = functools.partial(
            output_to_db,
            subregion_key=subregion_key,
            output_path=output_path,
            accessed=accessed,
        )
    elif args.output_format == "csv":
        raise NotImplementedError("Only output to SQLite currently supported")
        output_f = functools.partial(
            output_to_csv,
            subregion_key=subregion_key,
            output_path=args.output,
            accessed=accessed,
        )
    else:
        raise NotImplementedError

    if args.input_format == ".csv":
        f = functools.partial(
            _split_from_csv,
            region_key=region_key,
            subregion_key=subregion_key,
            output_func=output_f,
        )
    else:
        raise NotImplementedError("Splitting from non-csv files not yet supported")

    with mp.Pool(args.threads) as p:
        p.map(f, files)

    added = api.add(args.survey_name, "catalog", output_path, overwrite=True)
    if added:
        print(f"Sucessfully built database for survey {args.survey_name}")
    else:
        print("Error: Dataset was sucessfully split but couldn't add it!")


def _split_from_csv(
    file_path: Path,
    region_key,
    subregion_key,
    output_func,
):
    print(f"Working on file {file_path}")
    df = pd.read_csv(file_path)
    try:
        region = df[region_key].unique()
    except (KeyError, AttributeError):
        print(file_path)
        return
    region_splits = {reg: df[df[region_key] == reg] for reg in region}
    output_func(region_splits)


def output_to_csv():
    pass


def output_to_db(data, subregion_key, output_path, accessed):
    con = sqlite3.Connection(output_path, timeout=30)
    for region_key, region_values in data.items():
        if type(region_values) == pd.DataFrame:
            try:
                already_accessed = accessed[region_key]
            except KeyError:
                already_accessed = False

            if not already_accessed:
                accessed[region_key] = True
            else:
                while True:
                    a = accessed[region_key]
                    if not a:
                        accessed[region_key] = True
                        break
            while True:
                try:
                    region_values.to_sql(str(region_key), con=con, if_exists="append")
                    break
                except sqlite3.OperationalError:
                    continue
            if accessed:
                accessed[region_key] = False
    return
